import json
import os
import ast

CONFIG_FILE_PATH = os.path.join(os.getcwd(), 'chattdd_config.json')


def write_to_file(content, filename):
    config = load_config()
    output_folder = config['OUTPUTFOLDER']
    file_path = os.path.join(output_folder, filename)
    
    # Ensure the directory exists
    directory_path = os.path.dirname(file_path)
    os.makedirs(directory_path, exist_ok=True)
    
    # Check if __init__.py exists in the directory
    init_file_path = os.path.join(directory_path, '__init__.py')
    if not os.path.exists(init_file_path):
        with open(init_file_path, 'w') as init_file:
            pass  # create an empty __init__.py file
    
    # Write the provided content to the specified file
    with open(file_path, 'w') as file:
        file.write(content)

def load_config():
    if not os.path.exists(CONFIG_FILE_PATH):
        # Create config file with default values if it doesn't exist
        default_config = {
            "CHATTDD_MODEL": "text-davinci-003",
            "OUTPUTFOLDER": "ChatTDDgenerated"
        }
        with open(CONFIG_FILE_PATH, 'w') as file:
            json.dump(default_config, file, indent=4)
        return default_config
    else:
        with open(CONFIG_FILE_PATH, 'r') as file:
            return json.load(file)


def update_config(key, value):
    config = load_config()
    config[key] = value
    with open(CONFIG_FILE_PATH, 'w') as file:
        json.dump(config, file, indent=4)

def is_valid_syntax(code_str):
    try:
        compile(code_str, '<string>', 'exec')
        return True
    except SyntaxError:
        return False

def parse_string_to_dict(input_str: str) -> dict:
    try:
        # First, attempt to convert the string into a Python object using ast.literal_eval
        result = ast.literal_eval(input_str)
        
        # Check if the result is a dictionary
        if isinstance(result, dict):
            # Trim whitespaces for string values
            for key, value in result.items():
                if isinstance(value, str):
                    result[key] = value.strip()
            return result

    except (ValueError, SyntaxError):
        pass

    try:
        # If ast.literal_eval fails, then try parsing as JSON
        result = json.loads(input_str)
        
        if isinstance(result, dict):
            # Trim whitespaces for string values
            for key, value in result.items():
                if isinstance(value, str):
                    result[key] = value.strip()
            return result

    except json.JSONDecodeError:
        return {"comment": input_str}

    return {"comment": input_str}




import click
import chattdd.logic

@click.group(invoke_without_command=True)
@click.pass_context
def cli(ctx):
    if ctx.invoked_subcommand is None:
        try:
            user_requirement = click.prompt("Please enter the functional requirement")
            ctx.invoke(test_and_code, user_requirement=(user_requirement,))
        except Exception as e:
            click.echo(f"An error occurred: {str(e)}")

@cli.command()
@click.argument('outputfolder')
def outputfolder(outputfolder):
    try:
        logic.set_output_folder(outputfolder)
        click.echo(f'Output folder set to {outputfolder}')
    except Exception as e:
        click.echo(f"Failed to set output folder. Error: {str(e)}")

@cli.command()
@click.argument('model_name', type=click.Choice(['text-davinci-003', 'gpt-3.5-turbo', 'gpt-4'], case_sensitive=False))
def model(model_name):
    try:
        logic.set_model(model_name)
        click.echo(f'Model set to {model_name}')
    except Exception as e:
        click.echo(f"Failed to set model. Error: {str(e)}")

@cli.command()
@click.argument('user_requirement', nargs=-1)
def test_and_code(user_requirement):
    user_requirement_str = ' '.join(user_requirement)
    comment, output = logic.generate_and_save_logic(user_requirement_str)
    if output:
        click.echo(f"\nGenerated test for function: {output['function_name']}")
        click.echo(f"..also generated function: {output['function_name']}")
    else:
        click.echo(comment)

@cli.command()
@click.argument('user_requirement', nargs=-1)
def test(user_requirement):
    user_requirement_str = ' '.join(user_requirement)
    comment, _ = logic.generate_and_save_logic(user_requirement_str, save_function_code=False)
    click.echo(comment)

if __name__ == "__main__":
    cli()



import os
import keyring
from langchain.chat_models import ChatOpenAI
from langchain.llms import OpenAI
from langchain.prompts import ChatPromptTemplate
from langchain.schema.output_parser import StrOutputParser
from langchain.globals import set_llm_cache
from langchain.cache import InMemoryCache
from chattdd.tools import load_config, parse_string_to_dict, is_valid_syntax


def initialize_model(model_name=None):
    config = load_config()
    model_name = config['CHATTDD_MODEL']

    openai_api_key = os.getenv('OPENAI_API_KEY')
    if not openai_api_key:
        openai_api_key = keyring.get_password('openai', 'api_key')
        if not openai_api_key:
            openai_api_key = input("Please enter your OpenAI API key: ")
            keyring.set_password('openai', 'api_key', openai_api_key)
    
    os.environ['OPENAI_API_KEY'] = openai_api_key
    set_llm_cache(InMemoryCache())

    if model_name == 'text-davinci-003':
        model = OpenAI(model_name="text-davinci-003", max_tokens=3000, temperature=0, openai_api_key=openai_api_key)
    elif model_name == 'gpt-3.5-turbo':
        model = ChatOpenAI(model_name="gpt-3.5-turbo", max_tokens=3000, temperature=0, openai_api_key=openai_api_key)
    elif model_name == 'gpt-4':
        model = ChatOpenAI(model_name="gpt-4", max_tokens=3000, temperature=0, openai_api_key=openai_api_key)
    else:
        raise ValueError(f'Unsupported model: {model_name}')
    
    return model


def generate_test_code(model, user_input, comment_from_last_run):
    config = load_config()
    base_directory = config['OUTPUTFOLDER']
    
    template = """
    You are a code generator that only can produce Python dictionaries. You follow all principles of Test Driven Development. 
    You adhere to PEP 8. 
    You will never use double or tripple quotes, in this case replace with the keyword '$%$'. 
    Your task is to generate function code AND pytest test code, and place it in a Python Dictionary return without any explanations or comments. 
    When writing the pytest file please include required setup or teardown code and use a descriptive test function name. 
    Both the test and function file will be saved in the base directory named {base_directory}, so add that to the include path of the test file.
    The user will prompt you with a required task. Your job is to write the function in one file and the pytest for such a function in another file. Remember to import the function name into the test file.
    {comment_from_last_run}
    
    No introduction or politeness. Your response MUST be a Python Dictionary in the following format:
     1. original_request: the original request from the user. No changes.
     2. function_name: a short function name that reflects the users request.
     3. test_code: python code file that includes a single function, namely the pytest for a function that would test a function that soleves the users requirements. The function to test will be found in the same directory as the test file and have a filename you suggest, so remember the "import from".
     4. test_file_name: file name for the pytest test code.
     5. pytest_result: an empty string (we will fill this later)
     6. function_code: python code file that includes a single function, namely the function that the pytest is testing. The function solves the users requirements. No comments.
     7. function_file_name: file name used in the test file.

    """
    generated_test_code_prompt = ChatPromptTemplate.from_messages(
        [("system", template), ("human", f"{user_input}")]
    )
    chain = generated_test_code_prompt | model | StrOutputParser()

    generated_test_code_result = chain.invoke({'user_input': user_input, 'base_directory': base_directory, 'comment_from_last_run': comment_from_last_run})
    
    data_dict = parse_string_to_dict(generated_test_code_result)

    return data_dict

def review_test_code(model, original_request, function_name, test_code):
    template = """
    You are a Pytest and TDD expert that only can produce Python dictionaries. Your task is to evaluate a pytest file. The file should contain one or more to tests to evaluate a function {function_name}. The function {function_name} has the following requirement: {original_request}

    You don't have the {function_name} code, only the pytest code. This is correct and should not reject the test code based on that.
    Here's the pytest file contents represented in a string, it contains escape sequences that you should ignore: 
    Remember to check that {function_name} is imported. 

    "
    {test_code}
    "

    Reject the test if one or more of the following occur: 
    - There are syntax errors in the test code
    - The test isn't following best practice for TDD with Pytest. 
    - The test fails to test a function that solves what the user asked for, {original_request}

    No introduction or politeness. Your response MUST be a Python Dictionary in the following format:
    1. result: "GO" if the test can be accepted. "REJECTED" if the test isn't up your standards.
    2. comment: an evaluation of the test code and the test's overall performance for testing a function to {original_request}

    """
    generated_test_review_prompt = ChatPromptTemplate.from_messages(
        [("system", template)]
    )
    chain = generated_test_review_prompt | model | StrOutputParser()
    chain_test_review = chain.invoke({'original_request': original_request, 'function_name': function_name, 'test_code': test_code})
    data_dict = parse_string_to_dict(chain_test_review)

    return data_dict


if __name__ == "__main__":
    from chattdd.tools import write_to_file
    import pytest

    model = initialize_model()
    #user_input = "sort a list alphabetically descending order"
    #user_input = "sort a list of objects alphabetically"
    #user_input = "list the prime numbers between 100 and 1000"
    #user_input = "Please implement a function that sorts a list alphabetically "
    user_input = 'test this' #should fail code generation
    test_code_dict = generate_test_code(model, user_input, )
    comment = None
    try:
        original_request = test_code_dict['original_request']
        function_name = test_code_dict['function_name']
        test_code = test_code_dict['test_code']
        test_file_name = test_code_dict['test_file_name']
        function_file_name = test_code_dict['function_file_name']
        function_code = test_code_dict['function_code']
    except KeyError as e:
        comment = test_code_dict['comment']
        print(comment)
        raise KeyError(f"ChatTDD made a mistake. The key {e} not found in the provided dictionary.")
    test_review_json = review_test_code(model, original_request, function_name, test_code)
    
    if is_valid_syntax(test_code):
        write_to_file(test_code, test_file_name)
    else:
        print("Test code is invalid")
    
    if is_valid_syntax(function_code):
        write_to_file(function_code, function_file_name)
    else:
        print("Function code is invalid")
    
    pytest.main(['-v'])



import pytest
from chattdd.core import initialize_model, generate_test_code, review_test_code
from chattdd.tools import update_config, load_config, write_to_file

def set_output_folder(outputfolder):
    update_config('OUTPUTFOLDER', outputfolder)

def set_model(model_name):
    update_config('CHATTDD_MODEL', model_name)

def generate_and_save_logic(user_requirement, save_function_code=True, iterations=3):
    config = load_config()
    model_name = config['CHATTDD_MODEL']
    model = initialize_model(model_name)
    
    comment = ''

    for i in range(iterations):
        generated_test_code_output = generate_test_code(model, user_requirement, comment)

        test_code = generated_test_code_output.get('test_code')
        if not test_code:
            return generated_test_code_output.get('comment', 'No comment available'), None

        review_output = review_test_code(
            model=model,
            original_request=generated_test_code_output['original_request'],
            function_name=generated_test_code_output['function_name'],
            test_code=test_code
        )

        review_result = review_output.get('result')
        comment = review_output.get('comment', 'No comment available')

        if review_result == "GO":
            break

    if review_result == "GO":
        write_to_file(test_code, generated_test_code_output['test_file_name'])

        if save_function_code:
            write_to_file(generated_test_code_output['function_code'], generated_test_code_output['function_file_name'])
        
        pytest.main(['-v'])
        return comment, generated_test_code_output

    return comment, None



from setuptools import setup, find_packages

setup(
    name='ChatTDD',
    version='0.2.7',
    packages=find_packages(),
    install_requires=[
        'click',
        'openai>=0.27.0',
        'pytest',
        'python-dotenv',
        'langchain',
        'keyring'
    ],
    entry_points={
        'console_scripts': [
            'chattdd = chattdd.cli:cli',
        ],
    },
    author='Simen Huuse',
    author_email='s.huuse+chattdd@gmail.com',
    description='A utility to assist in building Test-Driven Python code using Langchain and OpenAI.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/shuuse/ChatTDD',
    license='MIT',
    classifiers=[
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License',
    ],
)
